@@ -189,6 +189,7 @@ io_channel_handler (GIOChannel * channel, GIOCondition cond, GNetworkUdpDatagram
       udp->_priv->sockfd = -1;
 
       udp->_priv->dgram_status = GNETWORK_DATAGRAM_CLOSED;
+      g_object_notify (G_OBJECT (udp), "socket");
       g_object_notify (G_OBJECT (udp), "status");
       return FALSE;
     }
@@ -243,6 +244,7 @@ io_channel_handler (GIOChannel * channel, GIOCondition cond, GNetworkUdpDatagram
 	      udp->_priv->sockfd = -1;
 
 	      udp->_priv->dgram_status = GNETWORK_DATAGRAM_CLOSED;
+      	      g_object_notify (G_OBJECT (udp), "socket");
 	      g_object_notify (G_OBJECT (udp), "status");
 	      retval = FALSE;
 	      break;
@@ -319,6 +321,7 @@ io_channel_handler (GIOChannel * channel, GIOCondition cond, GNetworkUdpDatagram
 		  udp->_priv->sockfd = -1;
 
 		  udp->_priv->dgram_status = GNETWORK_DATAGRAM_CLOSED;
+		  g_object_notify (G_OBJECT (udp), "socket");
 		  g_object_notify (G_OBJECT (udp), "status");
 		  retval = FALSE;
 		  break;
@@ -356,6 +359,7 @@ io_channel_handler (GIOChannel * channel, GIOCondition cond, GNetworkUdpDatagram
 	      udp->_priv->sockfd = -1;
 
 	      udp->_priv->dgram_status = GNETWORK_DATAGRAM_CLOSED;
+      	      g_object_notify (G_OBJECT (udp), "socket");
 	      g_object_notify (G_OBJECT (udp), "status");
 	      retval = FALSE;
 	    }
@@ -717,17 +721,19 @@ gnetwork_udp_datagram_close (GNetworkUdpDatagram * udp)
 
   if (udp->_priv->channel != NULL)
     {
+      g_io_channel_shutdown (udp->_priv->channel, FALSE, NULL);
       g_io_channel_unref (udp->_priv->channel);
+      udp->_priv->source_cond = 0;
+      udp->_priv->source_id = 0;
       udp->_priv->channel = NULL;
     }
-
-  if (udp->_priv->sockfd > 0)
+  else if (udp->_priv->sockfd > 0)
     {
       shutdown (udp->_priv->sockfd, SHUT_RDWR);
       close (udp->_priv->sockfd);
-      udp->_priv->sockfd = -1;
     }
 
+  udp->_priv->sockfd = -1;
   udp->_priv->dgram_status = GNETWORK_DATAGRAM_CLOSED;
   g_object_notify (G_OBJECT (udp), "socket");
   g_object_notify (G_OBJECT (udp), "status");
